# from pathlib import Path
# from pyeio.common.types import FilePath, PyTOML, SerializedTOML
# from pyeio.common.exceptions import MissingExtraError
# from pyeio.file import io, utils
# from pyeio.io import web

# try:
#     import toml
# except ImportError:
#     raise MissingExtraError(extra="toml")


# def parse(data: SerializedTOML) -> PyTOML:
#     if isinstance(data, bytes):
#         # todo.fix: will not always work
#         data = str(data.decode())
#     if not isinstance(data, str):
#         # todo: add details on type error
#         raise TypeError()
#     toml_data = toml.loads(data)
#     return toml_data


# def serialize(data: PyTOML) -> str:
#     serialized_data = toml.dumps(data)
#     return serialized_data


# def open(
#     path: FilePath,
#     validate_file_extension: bool = True,
# ) -> PyTOML:
#     path = Path(path)
#     if validate_file_extension:
#         utils.run_file_extension_validation(
#             module_name="toml",
#             path=path,
#         )
#     serialized_toml = io.open_text(path=path)
#     return parse(serialized_toml)


# def save(
#     data: PyTOML,
#     path: FilePath,
#     overwrite: bool = False,
#     validate_file_extension: bool = True,
# ) -> None:
#     """
#     Save data to a TOML file.

#     Args:
#         data (PyTOML): TOML-compatible Python data structure to serialize and save.
#         path (FilePath): The path to the file as a `str` or `pathlib.Path`.
#         overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
#         validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
#     """
#     path = Path(path)
#     if validate_file_extension:
#         utils.run_file_extension_validation(
#             module_name="toml",
#             path=path,
#         )
#     serialized_toml: str = serialize(data)
#     io.save_text(data=serialized_toml, path=path, overwrite=overwrite)


# def load(
#     url: str,
#     chunk_size: int = 1 << 10,
#     follow_redirects: bool = True,
#     evaluate_size: bool | None = None,
#     show_progress: bool = False,
#     show_file_name: bool = True,
# ) -> PyTOML:
#     """
#     Load a TOML file from a URL directly into memory.

#     Args:
#         url (str): The URL of the TOML file.
#         chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
#         follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
#         evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
#         show_progress (bool, optional): Show progress bar. Defaults to False.
#         show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

#     Returns:
#         PyTOML: TOML data as built-in Python data structures.
#     """
#     if evaluate_size is None:
#         evaluate_size = show_progress
#     binary_data = web.load_binary(
#         url=url,
#         chunk_size=chunk_size,
#         follow_redirects=follow_redirects,
#         evaluate_size=evaluate_size,
#         show_progress=show_progress,
#         show_file_name=show_file_name,
#     )
#     serialized_data = binary_data.decode("utf-8")
#     toml_data = parse(serialized_data)
#     return toml_data


# def download(
#     url: str,
#     path: FilePath,
#     overwrite: bool = False,
#     validate_file_extension: bool = True,
#     chunk_size: int = 1 << 10,
#     follow_redirects: bool = True,
#     evaluate_size: bool | None = None,
#     show_progress: bool = False,
#     show_file_name: bool = True,
# ) -> None:
#     """
#     Download a TOML file from a URL directly to a local TOML file on disk.

#     Args:
#         url (str): The URL of the TOML file.
#         path (FilePath): The path to the file as a `str` or `pathlib.Path`.
#         overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
#         validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
#         chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
#         follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
#         evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
#         show_progress (bool, optional): Show progress bar. Defaults to False.
#         show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

#     Raises:
#         FileExistsError: File already exists. Set `overwrite` to `True` to bypass.
#     """
#     if evaluate_size is None:
#         evaluate_size = show_progress
#     path = Path(path)
#     if validate_file_extension:
#         utils.run_file_extension_validation(
#             module_name="toml",
#             path=path,
#         )
#     if path.exists() and (not overwrite):
#         raise FileExistsError(str(path))
#     toml_data = load(
#         url=url,
#         chunk_size=chunk_size,
#         follow_redirects=follow_redirects,
#         evaluate_size=evaluate_size,
#         show_progress=show_progress,
#         show_file_name=show_file_name,
#     )
#     save(
#         data=toml_data,
#         path=path,
#         overwrite=overwrite,
#         validate_file_extension=validate_file_extension,
#     )



# This file is unrelated to PyPy - it's just for profiling differences in speed of same underlying functionality.

from pydantic import PositiveInt
from pyeio.core.types import FilePath


def count_lines_in_file(path: FilePath, chunk_size: PositiveInt = 1 << 20) -> int:
    """
    Count the number of lines in a file.

    Args:
        path (FilePath): Path to the file.
        chunk_size (PositiveInt, optional): Size of chunks to read in bytes. Defaults to 1<<20.

    Returns:
        int: Number of lines in the file.
    """

    lines = 0
    with open(path, "rb") as f:
        read_chunk = f.read
        chunk = read_chunk(chunk_size)
        while chunk:
            lines += chunk.count(b"\n")
            chunk = read_chunk(chunk_size)
    return lines



import psutil
from pydantic import PositiveInt
from pathlib import Path
from pyeio.core.types import FilePath
from . import py

# todo: add automatic fallback to py functions if rust is unavailable
try:
    from . import rs
except ImportError:
    # todo: replace with custom exception
    raise Exception(
        "Missing rust binaries. Please submit a GitHub issue and I'll try and get to the bottom of this."
    )


def count_lines_in_file(
    path: FilePath,
    chunk_size: PositiveInt = 1 << 20,
    num_threads: PositiveInt | None = None,
) -> int:
    """
    Count the number of lines in a file.

    Args:
        path (FilePath): Path to the file.
        chunk_size (PositiveInt, optional): Size of chunks to read in bytes. Defaults to 1<<20.
        num_threads (PositiveInt | None, optional): Number of threads to use. If `None`, automatically determined.

    Returns:
        int: _description_
    """
    path = str(path)
    if not Path(path).exists():
        raise FileNotFoundError(path)
    if chunk_size < 1:
        raise TypeError(f"'chunk_size' must be a positive integer, is: {chunk_size}")
    num_available_threads = psutil.cpu_count(logical=True)
    if num_threads is None:
        num_threads = int(num_available_threads / 1.5)
    else:
        if num_threads < 1:
            raise TypeError(
                f"'num_threads' must be a positive integer, is: {num_threads}"
            )
        elif num_threads > num_available_threads:
            raise ValueError(
                f"Requested more threads than available ({num_threads} > {num_available_threads})"
            )
    return rs.count_lines_in_file(path, chunk_size, num_threads)
